#include "cell-pop.hpp"

CellPop::CellPop():
  nomultithread(Para::n_cpu == 1),
  n_alive_cells(0)
{}

void CellPop::reset(std::size_t n_cell)
{
  /* reset_parallel() is not faster */
  reset_serial(n_cell);
}

void CellPop::reset_serial(std::size_t n_cell)
{
  /* Add a Cell if there are not sufficient Cells */
  if(cellvec.real_size() < n_cell){
    /* In C++11, new elements are default constructed */
    cellvec.resize(n_cell);
  }

  /* Clear Molecules in all Cells. We need not clear all Cells; we only
     have to clear n_cell Cells. */
  for(std::size_t i=0; i < n_cell; ++i){
    cellvec[i].clear();
  }

  n_alive_cells = n_cell;
}

class CellPop::Reset {
  CellVec& cellvec;
public:
  void operator()(const tbb::blocked_range<std::size_t>& range) const {
    for(std::size_t index=range.begin(), end=range.end(); index != end; ++index){    
      cellvec[index].clear();
    }
  }
  Reset(CellVec& _cellvec) :
    cellvec(_cellvec)
  {}
};

void CellPop::reset_parallel(std::size_t n_cell)
{
  /* Add a Cell if there are not sufficient Cells */
  if(cellvec.real_size() < n_cell){
    /* In C++11, new elements are default constructed */
    cellvec.resize(n_cell);
  }

  /* Clear Molecules in all Cells. We need not clear all Cells; we only
     have to clear n_cell Cells. */
  tbb::parallel_for(tbb::blocked_range<std::size_t>(0, n_cell), Reset(cellvec), tbb::static_partitioner());

  n_alive_cells = n_cell;
}

void CellPop::add_mol(std::size_t cell_index, Molecule* mol_ptr)
{
  /* Bound check */
  try{
    Assert<GeneralError>(!ASSERT::ERROR_CHECK|| (cell_index < n_alive_cells));
  }
  catch(GeneralError){
    //    ParallelPrint::cerr("CellPop::add_mol() ERROR index out of bound\n");
    exit(-1);
  }

  cellvec[cell_index].add_mol(mol_ptr);
}

std::size_t CellPop::cell_size(std::size_t cell_index) const
{
  /* Bound check */
  try{
    Assert<GeneralError>(!ASSERT::ERROR_CHECK || cell_index < n_alive_cells);
  }
  catch(GeneralError){
    //    ParallelPrint::cerr("CellPop::cell_size() ERROR index out of bound\n");
    exit(-1);
  }
  
  return cellvec[cell_index].size();
}

std::size_t CellPop::get_n_alive_cells() const
{
  return n_alive_cells;
}

double CellPop::calc_fitness()
{
  /*debug*/
  // double parallel = calc_fitness_parallel();
  // double serial = calc_fitness_serial();
      
  // std::cout << serial - parallel << std::endl;
  // return serial;
  
  if(nomultithread){
    return calc_fitness_serial();
  }
  else{
    return calc_fitness_parallel();
  }
}

double CellPop::calc_fitness_serial()
{
  double local_sum_fitness = 0.0;

  for(std::size_t i=0, end=n_alive_cells; i < end; ++i){

    cellvec[i].calc_fitness();
    local_sum_fitness += cellvec[i].get_sum_fitness();
  }

  sum_fitness = local_sum_fitness;
  return local_sum_fitness;
}

class CellPop::CalculateFitness {
  CellVec& my_cellvec;
public:
  double my_sum_fitness;
  void operator()(const tbb::blocked_range<std::size_t>& range) {

    /* Use locals to speed up */
    double sum_fitness = my_sum_fitness;

    /* Go through Vesicles */
    for(std::size_t cell_index=range.begin(), end=range.end(); cell_index != end; ++cell_index){
      my_cellvec[cell_index].calc_fitness();
      sum_fitness += my_cellvec[cell_index].get_sum_fitness();
    }

    /* Don't forget to register the result in a non-local */
    my_sum_fitness = sum_fitness;
  }

  /* Constructor */
  CalculateFitness(CellVec& cellvec) :
    my_cellvec(cellvec),
    my_sum_fitness(0.0)
  {}

  /* Split constructor, which is used for parallel_reduce */
  CalculateFitness(CalculateFitness& x, tbb::split) :
    my_cellvec(x.my_cellvec),
    my_sum_fitness(0.0)
  {}

  /* A function used for parallel_reduce */
  void join(const CalculateFitness& y){my_sum_fitness += y.my_sum_fitness;}
};

double CellPop::calc_fitness_parallel()
{
  CalculateFitness calc_fitness(cellvec);
  tbb::parallel_reduce(tbb::blocked_range<std::size_t>(0, n_alive_cells), calc_fitness, tbb::static_partitioner());

  sum_fitness = calc_fitness.my_sum_fitness;
  return calc_fitness.my_sum_fitness;
}

void CellPop::add_nextgen_size(std::size_t cell_index)
{
  /* Bound check */
  try{
    Assert<GeneralError>(!ASSERT::ERROR_CHECK|| (cell_index < n_alive_cells));
  }
  catch(GeneralError){
    ParallelPrint::cerr("CellPop::add_nextgen_size() ERROR index out of bound\n");
    exit(-1);
  }

  cellvec[cell_index].add_nextgen_size();
}

unsigned long CellPop::get_nextgen_size(std::size_t cell_index) const
{
  /* Bound check */
  try{
    Assert<GeneralError>(!ASSERT::ERROR_CHECK|| (cell_index < n_alive_cells));
  }
  catch(GeneralError){
    ParallelPrint::cerr("CellPop::get_nextgen_size() ERROR index out of bound\n");
    exit(-1);
  }

  return cellvec[cell_index].get_nextgen_size();
}

void CellPop::set_parent_index(std::size_t cell_index, std::size_t parent_index)
{
  /* Bound check */
  try{
    Assert<GeneralError>(!ASSERT::ERROR_CHECK|| (cell_index < n_alive_cells));
  }
  catch(GeneralError){
    ParallelPrint::cerr("CellPop::set_parent_index() ERROR index out of bound\n");
    exit(-1);
  }

  cellvec[cell_index].set_parent_index(parent_index);
}

std::size_t CellPop::get_parent_index(std::size_t cell_index) const
{
  /* Bound check */
  try{
    Assert<GeneralError>(!ASSERT::ERROR_CHECK|| (cell_index < n_alive_cells));
  }
  catch(GeneralError){
    ParallelPrint::cerr("CellPop::get_parent_index() ERROR index out of bound\n");
    exit(-1);
  }

  return cellvec[cell_index].get_parent_index();
}

const std::vector<std::size_t>& CellPop::get_daughter_index(std::size_t cell_index) const
{
  /* Bound check */
  try{
    Assert<GeneralError>(!ASSERT::ERROR_CHECK|| (cell_index < n_alive_cells));
  }
  catch(GeneralError){
    ParallelPrint::cerr("CellPop::get_daughter_index() ERROR index out of bound\n");
    exit(-1);
  }

  return cellvec[cell_index].get_daughter_index();
}

/* This updates the daughter_index of Cells. It returns the total number
   of Cells in the next generation. */
std::size_t CellPop::set_daughter_cell_index()
{
  /* Counter for the index of a next-generation Cell */
  std::size_t index_of_nextgen_cell = 0;

  /* Number of daughter Cells a given Cell will have */
  std::size_t n_daughter_cells, nextgen_size;

  /* Alias to speed up */
  const unsigned long division_threshold = Para::division_threshold;

  /* For each Cell in the current generation */
  for(std::size_t i=0, end=n_alive_cells; i < end; ++i){
    Cell& cell = cellvec[i];
    /* Get the number of Molecules in the next generation before cell
       division */
    nextgen_size = cell.get_nextgen_size();

    /* Calculate how many daughter Cells this Cell will have in the
       next generation */
    n_daughter_cells = nextgen_size / division_threshold + ((nextgen_size % division_threshold)? 1 : 0);
    
    /* Add the indexes of the daughter Cells. Note that n_daughter_cells
       can be 0. */
    for(std::size_t counter=0; counter < n_daughter_cells; ++counter){
      cell.add_daughter_index(index_of_nextgen_cell);
      ++index_of_nextgen_cell;
    }
  }

  /* Return the number of alive Cells in the next generation */
  return index_of_nextgen_cell;
}

void CellPop::check_consistency() const
{
  for(std::size_t i=0, end=n_alive_cells; i < end; ++i){
    cellvec[i].check_consistency(i);
  }
}

bool CellPop::has_molecule(std::size_t cell_index, const Molecule* mol_ptr) const
{
  /* Bound check */
  try{
    Assert<GeneralError>(!ASSERT::ERROR_CHECK|| (cell_index < n_alive_cells));
  }
  catch(GeneralError){
    ParallelPrint::cerr("CellPop::has_molecule() ERROR index out of bound\n");
    exit(-1);
  }

  return cellvec[cell_index].has_molecule(mol_ptr);
}

