// Standard Library
#include <vector>

// External Libraries
#include <tbb/tbb.h>
#include <tbb/tbb_thread.h>

// My Libraries
#include "assert.hpp"
#include "random_wrapper2.hpp"
#include "parallel-print.hpp"

// Internal 
#include "para.hpp"
#include "cell.hpp"
#include "molecule.hpp"

#ifndef CELL_POP
#define CELL_POP

class CellPop {
  friend class Observer;

public:
  CellPop();

  void reset(std::size_t n_cell);

  // void clear_mol();

  // /* Return the index of the first empty Cell */
  // std::size_t get_empty_index();

  /* Add a Molecule to the Cell at [index] */
  void add_mol(std::size_t cell_index, Molecule* mol_ptr);

  /* Calculate fitness of Molecules */
  double calc_fitness();

  std::size_t cell_size(std::size_t cell_index) const;

  std::size_t get_n_alive_cells() const;

  void add_nextgen_size(std::size_t cell_index);
  unsigned long get_nextgen_size(std::size_t cell_index) const;

  void set_parent_index(std::size_t cell_index, std::size_t parent_index);
  std::size_t get_parent_index(std::size_t cell_index) const;

  std::size_t set_daughter_cell_index();
  const std::vector<std::size_t>& get_daughter_index(std::size_t cell_index) const;

  /* Checks if mapping between Molecules and Cells are consistent. This
     is used for debugging. */
  void check_consistency() const;

  bool has_molecule(std::size_t cell_index, const Molecule* mol_ptr) const;

  /* Methods */
private:
  double calc_fitness_serial();
  double calc_fitness_parallel();
  void reset_serial(std::size_t n_cell);
  void reset_parallel(std::size_t n_cell);
  class Reset;
  class CalculateFitness;

  /* Data */
private:
  const bool nomultithread;

  class CellVec : public std::vector<Cell> {
    /* We make size() private because we must be careful with its
       use. The size of a CellVec is usually greater than the number of
       its valid elements, which is given by CellPop::n_alive_cells. It
       will cause a hard-to-detect bug if we accidentally loop though
       all the elements of a CellVec, which usually include invalid
       elements (I have made such a bag). To avoid introducing such a
       bug, we 'rename' size(). */
  private:
    using std::vector<Cell>::size;
  public:
    std::vector<Cell>::size_type real_size() const {return size();}
  };

  /* Data of Cells */
  CellVec cellvec;

  /* The total number of Cells. This is equal to or greater than
     cellvec.size(). I use this variable to count the number of Cells
     that contain at least one Molecules. The Cells that have no
     Molecules are not deconstructed, but stay in cellvec, in order to
     save time. Empty Cells are reused later. */
  std::size_t n_alive_cells;
  
  /* Sum of fitnesses, which is updated by calling calc_fitness() */
  double sum_fitness;
};

#endif
