#include "cell.hpp"

Cell::Cell():
  nomultithread(Para::n_cpu == 1),
  /* We reserve a sufficiently big space for molecules_ptr so that the
     space will never be exausted. We do not know beforehand how many
     elements will be created in molecules_prt. See also Cell::add_mol() */
  max_size((Para::single_cell_mode) ? (Para::n_molecule+1) : 3*Para::division_threshold),
  molecules_ptr(max_size,0),
  atomic_currgen_size(),
  nonatomic_currgen_size(0),
  atomic_nextgen_size(),
  nonatomic_nextgen_size(0),
  parent_index(0)
{}

/* Remove all Molecules */
void Cell::clear()
{
  /* The following line is not necessary, but might be useful because it
     ensures that "untouchable" elements point to an invalid address so
     that if an error occurs, it's easier to detect. But, for the sake
     of speed, it is commented out. */
  //std::fill_n(molecules_ptr.begin(), atomic_currgen_size, nullptr);

  if(nomultithread){
    nonatomic_currgen_size = 0;
    nonatomic_nextgen_size = 0;
  }
  else{
    atomic_currgen_size = 0;
    atomic_nextgen_size = 0;
  }

  daughter_index.clear();
}

/* Return the number of Molecules */
std::size_t Cell::size() const
{
  if(nomultithread){
    return nonatomic_currgen_size;
  }
  else{
    return atomic_currgen_size;
  }
}

std::size_t Cell::fetch_and_add_currgen_size()
{
  if(nomultithread){
    return nonatomic_currgen_size++;
  }
  else{
    return atomic_currgen_size.fetch_and_add(1);
  }
}

/* Add a Molecule */
void Cell::add_mol(Molecule* mol_ptr)
{
  std::size_t index = fetch_and_add_currgen_size();

  /* Error check */
  try{
    Assert<GeneralError>(!ASSERT::ERROR_CHECK|| index < max_size);
  }
  catch(GeneralError){
    ParallelPrint::cerr("Cell::add_mol() ERROR: The number of Molecules exceeded max_size\n");
    exit(-1);
  }
  
  /* Add the pointer to a Molecule to this Cell */
  molecules_ptr[index] = mol_ptr;
}

void Cell::calc_average_k(double& ave_k) const
{
  double sum_k = 0.0;
  unsigned long local_n = 0;

  for(std::size_t ind=0, end=size(); ind<end; ++ind){
    const Molecule& mol =  *molecules_ptr[ind];

    sum_k += mol.get_k();
    ++local_n;
  }

  if(local_n){
    ave_k = sum_k/local_n;
  }
  else{
    ave_k = Para::min_k;
  }
}

void Cell::calc_average_exp_k(double& ave_k, double& ave_exp_sk)
{
  double sum_k = 0.0, sum_exp_sk = 0.0;
  unsigned long local_n = 0;
  const double s_mol = Para::selection_molecular;

  for(std::size_t ind=0, end=size(); ind<end; ++ind){
    const Molecule& mol =  *molecules_ptr[ind];

    sum_k += mol.get_k();
    sum_exp_sk += exp_cash(s_mol * mol.get_k());
    ++local_n;
  }

  if(local_n){
    ave_k = sum_k/local_n;
    ave_exp_sk = sum_exp_sk/local_n;
  }
  else{
    ave_k = Para::min_k;
    ave_exp_sk = exp_cash(s_mol * Para::min_k);
  }
}

/* Calculate the fitness of molecules */
void Cell::calc_fitness()
{
  /* Check if we have to kill a small Cell */
  if(!kill_small_cell()){
    /* Calculate the fitnesses of this Molecule */
    switch(Para::fitness_type){
    case Para::FitnessType::Neutral:
      calc_fitness_Neutral();
      break;
    case Para::FitnessType::Exp_Yukawa:
      calc_fitness_Exp_Yukawa();
      break;
    case Para::FitnessType::Lin_Yukawa:
      calc_fitness_Lin_Yukawa();
      break;
    case Para::FitnessType::Exp_Simple_MolecularNeutral:
      calc_fitness_Exp_Simple_MolecularNeutral();
      break;
    case Para::FitnessType::Exp_Simple:
      calc_fitness_Exp_Simple();
      break;
    default:
      ParallelPrint::cerr("Cell::calc_fitness() ERROR! Unkown value in Para::fitness_type\n");
      exit(-1);
    }

    /* Clear the dictionary as soon as possible */
    exp_map.clear();
  }
}

void Cell::calc_fitness_Neutral()
{
  /* Set the fitness of molecules */
  for(std::size_t ind=0, end=size(); ind<end; ++ind){
    Molecule& mol =  *molecules_ptr[ind];
    mol.set_fitness(1.0);
  }
  sum_fitness = molecules_ptr.size();
}

void Cell::calc_fitness_Exp_Yukawa()
{
  /* Calculate the sum of k values */
  double ave_k;
  calc_average_k(ave_k);

  /* Calculate the fitness of molecules */
  double fitness, sum=0.0;
  const double max_k = Para::max_k;
  
  for(std::size_t ind=0, end=size(); ind<end; ++ind){
    Molecule& mol =  *molecules_ptr[ind];

    fitness = exp_cash(ave_k*(max_k - mol.get_k()));
    mol.set_fitness(fitness);
    sum += fitness;
  }
  sum_fitness = sum;
}

void Cell::calc_fitness_Lin_Yukawa()
{
  /* Calculate the sum of k values */
  double ave_k;
  calc_average_k(ave_k);

  /* Calculate the fitness of molecules */
  double fitness, sum=0.0;
  const double max_k = Para::max_k;
  
  for(std::size_t ind=0, end=size(); ind<end; ++ind){
    Molecule& mol =  *molecules_ptr[ind];

    fitness = ave_k*(max_k - mol.get_k());
    mol.set_fitness(fitness);
    sum += fitness;
  }
  sum_fitness = sum;
}

void Cell::calc_fitness_Exp_Simple_MolecularNeutral()
{
  /* Calculate the sum of k values */
  double ave_k;
  calc_average_k(ave_k);

  /* Calculate the fitness of molecules */
  double fitness, sum=0.0;
  
  for(std::size_t ind=0, end=size(); ind<end; ++ind){
    Molecule& mol =  *molecules_ptr[ind];

    fitness = exp_cash(Para::selection_cellular*ave_k);
    mol.set_fitness(fitness);
    sum += fitness;
  }
  sum_fitness = sum;
}

void Cell::calc_fitness_Exp_Simple()
{
  /* Calculate the sum of k values */
  double ave_k, ave_exp_sk;
  calc_average_exp_k(ave_k, ave_exp_sk);

  /* Calculate the fitness of molecules */
  double fitness, sum=0.0;
  const double s_mol = Para::selection_molecular;
  const double factor = exp_cash(Para::selection_cellular*ave_k)/ave_exp_sk;
    
  for(std::size_t ind=0, end=size(); ind<end; ++ind){
    Molecule& mol =  *molecules_ptr[ind];

    fitness = factor * exp_cash(s_mol * mol.get_k());
    mol.set_fitness(fitness);
    sum += fitness;
  }
  sum_fitness = sum;
}

double Cell::get_sum_fitness() const
{
  return sum_fitness;
}

void Cell::add_nextgen_size()
{
  if(nomultithread){
    ++nonatomic_nextgen_size;
  }
  else{
    atomic_nextgen_size.fetch_and_add(1);
  }
}

unsigned long Cell::get_nextgen_size() const
{
  if(nomultithread){
    return nonatomic_nextgen_size;
  }
  else{
    return atomic_nextgen_size;
  }
}

void Cell::set_parent_index(std::size_t index)
{
  parent_index = index;
}

std::size_t Cell::get_parent_index() const
{
  return parent_index;
}

bool Cell::check_consistency(std::size_t cell_index) const
{
  for(std::size_t ind=0, end=size(); ind<end; ++ind){
    const Molecule& mol =  *molecules_ptr[ind];

    if(cell_index != mol.get_cell_index()){
      ParallelPrint::cerr("Cell::check_consistency() ERROR: Molecules in this Cell do not have correct cell_index\n");
      return false;
    }
  }
 
  return true;
}

bool Cell::has_molecule(const Molecule* mol_ptr) const
{
  unsigned long count=0;
  for(std::size_t ind=0, end=size(); ind<end; ++ind){
    if(molecules_ptr[ind] == mol_ptr){
      ++count;
    }
  }

  if(count == 1){
    return true;
  }
  else if(count == 0){
    return false;
  }
  else{
    ParallelPrint::cerr("Cell::has_molecule() ERROR: The same Molecule occurs more than once\n");
    exit(-1);
  }
}

double Cell::exp_cash(double x)
{
  auto it = exp_map.find(x);

  if(it != exp_map.end()){
    return it->second;
  }
  else{
    double y = exp(x);
    exp_map.insert(std::pair<double,double>(x, y));
    return y;
  }
}

void Cell::add_daughter_index(std::size_t index)
{
  daughter_index.push_back(index);
}

const std::vector<std::size_t>& Cell::get_daughter_index() const
{
  return daughter_index;
}

bool Cell::kill_small_cell()
{
  if(Para::kill_small_cell){
    /* If this Cell is too small */
    if(size() < Para::kill_threshold*Para::division_threshold){
      /* Set the fitness of Molecules to zero */
      for(std::size_t ind=0, end=size(); ind<end; ++ind){
	Molecule& mol =  *molecules_ptr[ind];
	mol.set_fitness(0.0);
      }
      sum_fitness = 0.0;

      /* Report that this Cell is killed */
      return true;
    }
    else{
      /* Report that this Cell is not killed */
      return false;
    }
  }
  else{
    /* Report that this Cell is not killed */
    return false;
  }
}
