// Standard Library
#include <vector>
#include <cmath>
#include <map>

// External Libraries
#include <tbb/tbb.h>
#include <tbb/tbb_thread.h>
#include <tbb/concurrent_vector.h>
#include <boost/dynamic_bitset.hpp>

// My Libraries
#include "assert.hpp"
#include "random_wrapper2.hpp"
#include "parallel-print.hpp"

// Internal 
#include "para.hpp"
#include "molecule.hpp"

#ifndef CELL
#define CELL

class Cell{
  friend class Observer;

public:
  Cell();

  std::size_t size() const;
  double get_sum_fitness() const;

  /* This is thread safe */
  void add_mol(Molecule* mol_ptr);

  /* Clears molecules, nextgen_size, daughter_index. */
  void clear();

  void calc_fitness();

  void add_nextgen_size();
  unsigned long get_nextgen_size() const;

  void set_parent_index(std::size_t index);
  std::size_t get_parent_index() const;

  void add_daughter_index(std::size_t index);
  const std::vector<std::size_t>& get_daughter_index() const;

  /* Checks if mapping between Molecules and Cells are consistent. This
     is used for debugging. Returns true if consistent. */
  bool check_consistency(std::size_t index) const;
  
  /* Returns true if this Cell has a given Molecule. */
  bool has_molecule(const Molecule* mol_ptr) const;

  bool kill_small_cell();

private:
  void calc_average_k(double& ave_k) const;
  void calc_average_exp_k(double& ave_k, double& ave_exp_sk);

  void calc_fitness_Exp_Yukawa();
  void calc_fitness_Lin_Yukawa();
  void calc_fitness_Neutral();
  void calc_fitness_Exp_Simple();
  void calc_fitness_Exp_Simple_MolecularNeutral();

  double exp_cash(double x);

  std::size_t fetch_and_add_currgen_size();
  
private:
  const bool nomultithread;
  const std::size_t max_size;

  /* Note that pointers to Molecules will not be invalidated because
     the number of Molecules is kept constant throughout a simulation */
  std::vector<Molecule*> molecules_ptr;

  /* The number of Molecules in the "current" generation. This DIFFES
     from molecules_ptr->size(). The latter set to a sufficiently large
     number, so that it is always greater than the former. */
  tbb::atomic<std::size_t> atomic_currgen_size;
  std::size_t nonatomic_currgen_size;

  /* The number of Molecules in the "next" generation. */
  tbb::atomic<std::size_t> atomic_nextgen_size;
  std::size_t nonatomic_nextgen_size;

  double sum_fitness;

  /* The index of the Cell from which this Cell has descended. */
  std::size_t parent_index;

  /* Indexes of the next-generation Cells that are the daughters of this
     Cell */
  std::vector<std::size_t> daughter_index;

  /* This is used to cash std::exp() */
  std::map<double, double> exp_map;

};

#endif
