// Libraries
#include <cmath>

// My Libraries
#include "assert.hpp"
#include "cellular-automata.hpp"

// Internal 
#include "para.hpp"

#ifndef COVMAT
#define COVMAT

//enum Trait {X, Y};

/* This is a base class used to define covariance matrix classes. It
   contains data member and methods that calculate averages. */
template <unsigned long NTRAIT>
class CovMatBase {
public:
  CovMatBase();

  /* Stupid old c++ does not allow vector.resize() to use default
     constructor. I must supply a copy constructor, which does exactly
     the same as the default constructor */
  CovMatBase(const CovMatBase& rhs);

  /* std::vector() forces me to supply an assignment operator, which
     DOES NOTHING! */
  CovMatBase& operator=(const CovMatBase& rhs){return *this;}

  /* Return averages, (co)variances, and sample size */
  double ave(std::size_t trait) const;
  double min(std::size_t trait) const;
  double max(std::size_t trait) const;
  double cov(std::size_t trait0, std::size_t trait1) const;
  unsigned long sample_size() const;

  /* This sets all internal variables to zero. */
  void reset();

  /* member method */
protected:
  /* We use only a half of each matrix. This is a function to access
     the right elements of a matrix. */
  double& get_elm(CA2D<double>& data, std::size_t trait0, std::size_t trait1);
  double get_elm2(const CA2D<double>& data, std::size_t trait0, std::size_t trait1) const;
  
  /* member variables */
protected:
  /* The number of traits */
  unsigned long ntrait;

  /* This sets whether it can accept more samples */
  bool waiting_for_more_sample;

  /* Number of samples */
  unsigned long nsample; 

  /* Average of trait values */
  std::vector<double> _average;

  /* The minimum of trait values */
  std::vector<double> _min;

  /* The maximum of trait values */
  std::vector<double> _max;

  /* Covariance matrix of trait values */
  CA2D<double> _covariance;

  /* Sum of the trait values or sum of the average of the trait values
     depending on whether we are sampling actual entities (Molecules or
     Vesicles) or CovMatBase. */
  std::vector<double> _sum1D;

  /* Sum of squares of the trait values (x*x, x*y, y*y, etc.) or sum of
     the (co)variances depening on whether we are sampling actual
     entities (Molecules or Vesicles) or CovMatBase. */
  CA2D<double> _sum2D;
};

/* This is a "direct" covariance matrix class. It takes values of P_P2Q,
    P_Q2P, and so forth as samples calculate the covariances. */
template <unsigned long NTRAIT>
class CovMatDirect : public CovMatBase<NTRAIT> {
public:
  /* Add sample */
  void add_sample(const std::vector<double>& data, unsigned long n=1);

  /* Calculate averages and (co)variances. */
  void calc();

  /* "Standardize" (co)variances by dividing with averages */
  void standardize_variance();

private:
  /* Apparently, you need these to access members of the base class
     template. */
  using CovMatBase<NTRAIT>::ntrait;
  using CovMatBase<NTRAIT>::waiting_for_more_sample;
  using CovMatBase<NTRAIT>::nsample;
  using CovMatBase<NTRAIT>::_average;
  using CovMatBase<NTRAIT>::_min;
  using CovMatBase<NTRAIT>::_max;
  using CovMatBase<NTRAIT>::_covariance;
  using CovMatBase<NTRAIT>::_sum1D;
  using CovMatBase<NTRAIT>::_sum2D;
};

/* This is an average covariance matrix class. It takes covariance
   matrices as samples and calculate the average covariance matrices. */
template <unsigned long NTRAIT>
class CovMatAverage : public CovMatBase<NTRAIT> {
public:
  /* Add sample */
  void add_sample(const CovMatBase<NTRAIT>& matrix);

  /* Calculate averages and (co)variances. */
  void calc();

private:
  using CovMatBase<NTRAIT>::ntrait;
  using CovMatBase<NTRAIT>::waiting_for_more_sample;
  using CovMatBase<NTRAIT>::nsample;
  using CovMatBase<NTRAIT>::_average;
  using CovMatBase<NTRAIT>::_min;
  using CovMatBase<NTRAIT>::_max;
  using CovMatBase<NTRAIT>::_covariance;
  using CovMatBase<NTRAIT>::_sum1D;
  using CovMatBase<NTRAIT>::_sum2D;
};

template<unsigned long NTRAIT>
CovMatBase<NTRAIT>::CovMatBase(): 
  ntrait(NTRAIT),
  waiting_for_more_sample(true),
  nsample(0),
  _average(ntrait, 0.0),
  _min(ntrait, 0.0),
  _max(ntrait, 0.0),
  _covariance(ntrait,ntrait),
  _sum1D(ntrait, 0.0),
  _sum2D(ntrait,ntrait)
{
  reset();
}

/* If the NTRAIT of this constructor differs from the NTRAIT of rhs, the
   compiler will correctively gives error. */
template<unsigned long NTRAIT>
CovMatBase<NTRAIT>::CovMatBase(const CovMatBase<NTRAIT>& rhs): 
  ntrait(rhs.ntrait),
  waiting_for_more_sample(true),
  nsample(0),
  _average(ntrait, 0.0),
  _min(ntrait, 0.0),
  _max(ntrait, 0.0),
  _covariance(ntrait,ntrait),
  _sum1D(ntrait, 0.0),
  _sum2D(ntrait,ntrait)
{
  reset();
}

template<unsigned long NTRAIT>
void CovMatBase<NTRAIT>::reset()
{
  /* It can now accept new samples */
  waiting_for_more_sample = true;
  nsample = 0;

  /* Initialize variables */
  for(std::size_t i=0; i<ntrait; ++i){
    _sum1D[i] = 0.0;
    _min[i] =  std::numeric_limits<double>::max();
    _max[i] = -std::numeric_limits<double>::max();
    for(std::size_t j=i; j<ntrait; ++j){
      _sum2D.cell(i,j) = 0.0;
    }
  }
}

template<unsigned long NTRAIT>
double& CovMatBase<NTRAIT>::get_elm(CA2D<double>& data, std::size_t trait0, std::size_t trait1)
{
  /* We only use a half of each matrix */
  if(trait0 <= trait1){
    return data.cell(trait0, trait1);
  }else{
    return data.cell(trait1, trait0);
  }
}

template<unsigned long NTRAIT>
double CovMatBase<NTRAIT>::get_elm2(const CA2D<double>& data, std::size_t trait0, std::size_t trait1) const
{
  /* We only use a half of each matrix */
  if(trait0 <= trait1){
    return data.cell(trait0, trait1);
  }else{
    return data.cell(trait1, trait0);
  }
  return 0.;
}

template<unsigned long NTRAIT>
unsigned long CovMatBase<NTRAIT>::sample_size() const
{
  /* User cannot get covariance before calling calc() */
  if(waiting_for_more_sample){
    std::cerr << "CovMatBase::sample_size() ERROR you try to get a value before calling calc()." << std::endl;
    exit(-1);
  }
  else{
    return nsample;
  }
}

template<unsigned long NTRAIT>
double CovMatBase<NTRAIT>::ave(std::size_t trait) const
{
  /* User cannot get covariance before calling calc() */
  if(waiting_for_more_sample){
    std::cerr << "CovMatBase::ave() ERROR you try to get a value before calling calc()." << std::endl;
    exit(-1);
  }
  else{
    return _average[trait];
  }
}

template<unsigned long NTRAIT>
double CovMatBase<NTRAIT>::min(std::size_t trait) const
{
  /* User cannot get covariance before calling calc() */
  if(waiting_for_more_sample){
    std::cerr << "CovMatBase::ave() ERROR you try to get a value before calling calc()." << std::endl;
    exit(-1);
  }
  else{
    return _min[trait];
  }
}

template<unsigned long NTRAIT>
double CovMatBase<NTRAIT>::max(std::size_t trait) const
{
  /* User cannot get covariance before calling calc() */
  if(waiting_for_more_sample){
    std::cerr << "CovMatBase::ave() ERROR you try to get a value before calling calc()." << std::endl;
    exit(-1);
  }
  else{
    return _max[trait];
  }
}

template<unsigned long NTRAIT>
double CovMatBase<NTRAIT>::cov(std::size_t trait0, std::size_t trait1) const
{
  /* User cannot get covariance before calling calc() */
  if(waiting_for_more_sample){
    std::cerr << "CovMatBase::cov() ERROR you try to get a value before calling calc()." << std::endl;
    exit(-1);
  }
  else{
    return get_elm2(_covariance, trait0, trait1);
  }
}

template<unsigned long NTRAIT>
void CovMatDirect<NTRAIT>::add_sample(const std::vector<double>& data, unsigned long n)
{
  if(!waiting_for_more_sample){
    std::cerr << "CovMatDirect::add_sample() ERROR you try to add a sample before calling reset()." << std::endl;
    exit(-1);
  }

  if(n){
    nsample += n;

    /* Add samples */
    for(std::size_t i=0; i<ntrait; ++i){
      /* Calculate the sum of values */
      _sum1D[i] += data[i]*n;

      /* Calculate the maximum of values */
      if(_max[i] < data[i]){
	_max[i] = data[i];
      }

      /* Calculate the minimum of values */
      if(_min[i] > data[i]){
	_min[i] = data[i];
      }

      for(std::size_t j=i; j<ntrait; ++j){
	/* Calculate the sum of squared values */
	_sum2D.cell(i,j) += data[i]*data[j]*n;
      }
    }
  }
}

template<unsigned long NTRAIT>
void CovMatDirect<NTRAIT>::calc()
{
  if(waiting_for_more_sample && nsample){
    /* We are no mre accepting new samples (until reset() is called ) */
    waiting_for_more_sample = false;
  }
  else if(!waiting_for_more_sample){
    std::cerr << "CovMatDirect::calc() ERROR called twice?" << std::endl;
    exit(-1);
  }
  else if(nsample==0){
    /* When nsample is 0, we set _average and _covariance to 0 too. To
       this end, I set nsample to 1. Note that _sum1D and _sum2D are all
       0 at this point. */
    nsample = 1;
  }

  /* Calculate averages */
  for(std::size_t i=0; i<ntrait; ++i){
    _average[i] = _sum1D[i]/nsample;
    for(std::size_t j=i; j<ntrait; ++j){
      _covariance.cell(i,j) = (_sum2D.cell(i,j)-(_sum1D[i]*_sum1D[j])/nsample)/nsample;
    }
  }

  /* If waiting_for_more_sample==true at this point, nsample had been 0
     and was set to 1. */
  if(waiting_for_more_sample){
    /* Revert it to 0 */
    nsample = 0;

    /* We are no more accepting new samples (until reset() is called) */
    waiting_for_more_sample = false;
  }
}

template<unsigned long NTRAIT>
void CovMatDirect<NTRAIT>::standardize_variance()
{
  /* Calculate averages */
  for(std::size_t i=0; i<ntrait; ++i){
    _average[i] = _sum1D[i]/nsample;
    for(std::size_t j=i; j<ntrait; ++j){
      if(_average[i] * _average[j] != 0.0){
      _covariance.cell(i,j) /= (_average[i] * _average[j]);
      }
      else{
	std::cerr << "CovMatDirect::standardize_variance() ERROR: dividing by zero." << std::endl;
	exit(-1);
      }
    }
  }
}


template<unsigned long NTRAIT>
void CovMatAverage<NTRAIT>::add_sample(const CovMatBase<NTRAIT>& matrix)
{
  if(!waiting_for_more_sample){
    std::cerr << "CovMatAverage::add_sample() ERROR you try to add a sample before calling reset()." << std::endl;
    exit(-1);
  }

  /* The average of covariance matrices must reflect the size of each
     Cell. Thus, we increment nsample by the number of Molecules from
     which "matrix" is calculated. Each Molecule is considered to have
     the same covariance matrix within the same Cell.  */
  const unsigned long _nsample = matrix.sample_size();

  if(_nsample){
    nsample += _nsample;

    for(std::size_t i=0; i<ntrait; ++i){
      /* Calculate the sum of values */
      _sum1D[i] += matrix.ave(i)*_nsample;

      /* Calculate the maximum of values */
      if(_max[i] < matrix.ave(i)){
	_max[i] = matrix.ave(i);
      }

      /* Calculate the minimum of values */
      if(_min[i] > matrix.ave(i)){
	_min[i] = matrix.ave(i);
      }

      /* Calculate the sum of covariances */
      for(unsigned j=i; j<ntrait; ++j){
	_sum2D.cell(i,j) += matrix.cov(i,j)*_nsample;
      }
    }
  }
}

template<unsigned long NTRAIT>
void CovMatAverage<NTRAIT>::calc()
{
  if(waiting_for_more_sample && nsample){
    /* We are no more accepting new samples (until reset() is called) */
    waiting_for_more_sample = false;
  }
  else if(!waiting_for_more_sample){
    std::cerr << "CovMatAverage::calc() ERROR called twice?" << std::endl;
    exit(-1);
  }
  else if(nsample==0){
    /* When nsample is 0, we set _average and _covariance to 0 too. To
       this end, I set nsample to 1. Note that _sum1D and _sum2D are all
       0 at this point. */
    nsample = 1;
  }

  /* Calculate averages */
  for(std::size_t i=0; i<ntrait; ++i){
    _average[i] = _sum1D[i]/nsample;
    for(std::size_t j=i; j<ntrait; ++j){
      _covariance.cell(i,j) = _sum2D.cell(i,j)/nsample;
    }
  }

  /* If waiting_for_more_sample==true at this point, nsample had been 0
     and was set to 1. */
  if(waiting_for_more_sample){
    /* Revert it to 0 */
    nsample = 0;

    /* We are no more accepting new samples (until reset() is called) */
    waiting_for_more_sample = false;
  }
}

#endif
