#include "histogram.hpp"

Histogram::Histogram()
{
  reset();
}

void Histogram::reset()
{
  _min =  std::numeric_limits<double>::max();
  _max = -std::numeric_limits<double>::max();

  _index_of_min_bin = 0;
  _index_of_max_bin = 0;

  _mean = 0.0;
  _variance = 0.0;
  _skewness = 0.0;
  _kurtosis = 0.0;
  
  waiting_for_more_sample = true;

  if(!data.empty()){
    data.clear();
  }

  if(!_frequencies.empty()){
    _frequencies.clear();
  }
}

void Histogram::add_datum(double datum, unsigned long nsample)
{
  if(waiting_for_more_sample){
    if(nsample==1){
      data.push_back(datum);
    }
    else{
      data.insert(data.end(), nsample, datum);
    }

    /* Calculate the maximum of values */
    if(_max < datum){
      _max = datum;
    }

    /* Calculate the minimum of values */
    if(_min > datum){
      _min = datum;
    }
  }
  else{
    std::cerr << "Histogram::add_datum() ERROR calc() has already been called." << std::endl;
    exit(-1);
  }
}

double Histogram::bin_width() const
{
  /* This is just made up value */
  return Para::cell_stat_bin_width;
}

double Histogram::midpoint_of_min_bin() const
{
  if(waiting_for_more_sample){
    std::cerr << "Histogram::midpoint_of_min_bin() ERROR you try to get a value before calling calc()." << std::endl;
    exit(-1);
  }
  else{
    return bin_width() * _index_of_min_bin;
  }
}

long Histogram::index_of_min_bin() const
{
  if(waiting_for_more_sample){
    std::cerr << "Histogram::index_of_min_bin() ERROR you try to get a value before calling calc()." << std::endl;
    exit(-1);
  }
  else{
    return _index_of_min_bin;
  }
}

double Histogram::midpoint_of_max_bin() const
{
  if(waiting_for_more_sample){
    std::cerr << "Histogram::midpoint_of_max_bin() ERROR you try to get a value before calling calc()." << std::endl;
    exit(-1);
  }
  else{
    return bin_width() * _index_of_max_bin;
  }
}

long Histogram::index_of_max_bin() const
{
  if(waiting_for_more_sample){
    std::cerr << "Histogram::index_of_max_bin() ERROR you try to get a value before calling calc()." << std::endl;
    exit(-1);
  }
  else{
    return _index_of_max_bin;
  }
}

double Histogram::min() const
{
  if(waiting_for_more_sample){
    std::cerr << "Histogram::min() ERROR you try to get a value before calling calc()." << std::endl;
    exit(-1);
  }
  else{
    return _min;
  }
}

double Histogram::max() const
{
  if(waiting_for_more_sample){
    std::cerr << "Histogram::max() ERROR you try to get a value before calling calc()." << std::endl;
    exit(-1);
  }
  else{
    return _max;
  }
}

void Histogram::calc()
{
  /* New data is not accepted anymore */
  waiting_for_more_sample = false;

  /* If no data is given, do something special */
  if(data.empty()){
    return;
  }

  /* Calculate mean */
  _mean = gsl_stats_mean(&data[0],1,data.size());

  const double _bin_width = bin_width();

  /* When we make a histogram, we shift  */  
  const double shift = 0.5 * _bin_width;
  const double centralized_min = min() - _mean + shift;
  const double centralized_max = max() - _mean + shift;
  const long bin_index_offset = floor(centralized_min/_bin_width);

  _index_of_min_bin = bin_index_offset;
  _index_of_max_bin = floor(centralized_max/_bin_width);
  
  /* This is the total number of bins */
  const unsigned long n_bin = floor(centralized_max/_bin_width) - bin_index_offset + 1;

  /* Initialize _frequencies. If data.empty()==true, n_bin becomes an
     enourmous numbre because max() and min() are enourmous. In this
     case, assign() will through bad_alloc. */
  _frequencies.assign(n_bin, 0);

  for(auto& datum : data){
    /* Centralize data */
    datum -= _mean;
    
    /* Update frequency distribution */
    ++_frequencies[floor((datum + shift)/_bin_width) - bin_index_offset];
  }

  /* Calculate moments */
  double _sd;
  if(data.size() > 1){
    /* Variance */
    _variance = gsl_stats_variance_m(&data[0],1,data.size(), 0.0);

    /* Remove unbiasedness */
    _variance *= (data.size()-1)/static_cast<double>(data.size());
    _sd = sqrt(_variance);

    /* Other moments */
    double _3rd = 0.0, _4th = 0.0, tmp;
    for(auto& datum : data){
      tmp = datum * datum * datum;
      _3rd += tmp;
      tmp *= datum;
      _4th += tmp;
    }
    _central_moment_3rd = _3rd/static_cast<double>(data.size());
    _central_moment_4th = _4th/static_cast<double>(data.size());
  }
  else{
    _variance = 0.0;
    _sd = 0.0;
    _central_moment_3rd = 0.0;
    _central_moment_4th = 0.0;
  }

  /* Calculate indexes. */
  _skewness = gsl_stats_skew_m_sd(&data[0],1,data.size(), 0.0, _sd);
  _kurtosis = gsl_stats_kurtosis_m_sd(&data[0],1,data.size(), 0.0, _sd);
}

double Histogram::mean() const
{
  if(waiting_for_more_sample){
    std::cerr << "Histogram::mean() ERROR you try to get a value before calling calc()." << std::endl;
    exit(-1);
  }

  return _mean;
}

double Histogram::variance() const
{
  if(waiting_for_more_sample){
    std::cerr << "Histogram::variance() ERROR you try to get a value before calling calc()." << std::endl;
    exit(-1);
  }

  return _variance;
}

double Histogram::skewness() const
{
  if(waiting_for_more_sample){
    std::cerr << "Histogram::skewness() ERROR you try to get a value before calling calc()." << std::endl;
    exit(-1);
  }

  return _skewness;
}

double Histogram::excess_kurtosis() const
{
  if(waiting_for_more_sample){
    std::cerr << "Histogram::excess_kurtosis() ERROR you try to get a value before calling calc()." << std::endl;
    exit(-1);
  }

  return _kurtosis;
}

double Histogram::central_moment_3rd() const
{
  if(waiting_for_more_sample){
    std::cerr << "Histogram::central_moment_3rd() ERROR you try to get a value before calling calc()." << std::endl;
    exit(-1);
  }

  return _central_moment_3rd;
}

double Histogram::central_moment_4th() const
{
  if(waiting_for_more_sample){
    std::cerr << "Histogram::central_moment_4th() ERROR you try to get a value before calling calc()." << std::endl;
    exit(-1);
  }

  return _central_moment_4th;
}

unsigned long Histogram::sample_size() const
{
  return data.size();
}

const std::vector<unsigned long>& Histogram::frequencies() const
{
  return _frequencies;
}

