// Libraries
#include <cmath>
#include <algorithm>
#include <gsl/gsl_statistics_double.h>

// My Libraries
#include "assert.hpp"
#include "cellular-automata.hpp"

// Internal 
#include "para.hpp"

#ifndef HISTOGRAM
#define HISTOGRAM


class Histogram {
public:
  /* Constructors */
  Histogram();

  /* Return average, variance, and sample size, etc */
  double min() const;
  double max() const;
  double mean() const;
  double variance() const;
  double skewness() const;
  double excess_kurtosis() const;
  double central_moment_3rd() const;
  double central_moment_4th() const;
  unsigned long sample_size() const;

  /* *CENTRALIZED* frequency histogram. Values are first centralized
     (i.e., dX = X - <X>, where X is a real value) and then binned as
     follows:

       [x-w/2, x+w/2), [x+w/2, x+3w/2), [x+3w/2, x+5w/2), ...

     where x=bin_min() and w=bin_width(). The choice of bin_width() is
     arbitrary. The value of bin_min() is calculated such that the
     left-most bin, i.e. [x-w/2, x+w/2), contains the minimum value of
     data and that x=w*(integer). The latter condition implies that
     there is always a bin [-w/2, w/2), which will contain values close
     to 0 (i.e., values close to the mean).  */

  /* Returns w described above */
  double bin_width() const;

  /* Returns x described above */
  double midpoint_of_min_bin() const;

  /* This returns an integer i that satisfies x=w*i. i is usually
     negative. */
  long index_of_min_bin() const;

  /* Returns the mid-point of the bin that contains the maximum value */
  double midpoint_of_max_bin() const;

  /* This returns an integer i that satisfies
     w*i=midpoint_of_max_bin(). */
  long index_of_max_bin() const;

  const std::vector<unsigned long>& frequencies() const;

  /* Adding samples */
  void add_datum(double datum, unsigned long nsample=1);

  /* Calculate statistics. Must call this method before getting
     statistical information. */
  void calc();

  /* This sets all internal variables to zero. */
  void reset();

  
  /* member variables */
private:
  /* Raw data */
  std::vector<double> data;

  /* Frequency data */
  std::vector<unsigned long> _frequencies;

  /* The minimum of trait values */
  double _min;

  /* The maximum of trait values */
  double _max;

  long _index_of_min_bin;
  long _index_of_max_bin;
  
  double _mean;
  double _variance;
  double _central_moment_3rd;
  double _central_moment_4th;
  double _skewness;
  double _kurtosis;

  bool waiting_for_more_sample;
  
};


#endif
