#include "molecule.hpp"

Molecule::Molecule():
  cell_index(std::numeric_limits<std::size_t>::max())
{
  std::size_t n_loci_plus;
  
  switch(Para::mut_type){
  case Para::MutType::PerGenomeStepWrap:
  case Para::MutType::PerGenomeStepNegative:
  case Para::MutType::PerGenomeStepNoMinBoundary:
  case Para::MutType::PerGenomeStepNoMinMaxBoundary:
  case Para::MutType::PerGenomeStepWrapNormal:
  case Para::MutType::PerGenomeBinary:
    k = Para::init_k;
    break;
  case Para::MutType::PerGenomeMultiLocus:
    genome.assign(Para::n_loci, false);
    /* Initial number of loci with +1 allele */
    n_loci_plus = 0.5 + 0.5*(Para::init_k/Para::mutation_step + Para::n_loci);

    /* Error check */
    if(n_loci_plus >= genome.size()){
      ParallelPrint::cerr("Molecule::Molecule() ERROR! Para::init_k is too different from zero\n");
    }

    /* Set genome */
    for(std::size_t i=0; i<n_loci_plus; ++i){
      genome[i] = true;
    }
    /* Calculte k value */
    k = genome2k();
    break;
  default:
    ParallelPrint::cerr("Molecule::Molecule() ERROR! Unkown value in Para::muttype\n");
    exit(-1);
    break;
  }
}

void Molecule::set_k(double _k)
{
  k = _k;
}

void Molecule::set_cell_index(std::size_t index)
{
  cell_index = index;
}

std::size_t Molecule::get_cell_index() const
{
  return cell_index;
}

void Molecule::set_parent_index(std::size_t index)
{
  parent_index = index;
}

std::size_t Molecule::get_parent_index() const
{
  return parent_index;
}

void Molecule::mutate()
{
  RandomTBB2::reference random = random_tbb_2.local();
  mutate(random);
}

void Molecule::mutate(RandomTBB2::reference random)
{
  mutate_Full(random);
}

void Molecule::mutate_Full(RandomTBB2::reference random)
{
  if(random.Prob(Para::mutation_rate)){
    switch(Para::mut_type){
    case Para::MutType::PerGenomeStepWrap:
      change_parameter_step_wrap(k, Para::min_k, Para::max_k, Para::mutation_step, random);
      break;
    case Para::MutType::PerGenomeStepNegative:
    case Para::MutType::PerGenomeStepNoMinBoundary:
      change_parameter_negative(k, Para::max_k, Para::mutation_step, random);
      break;
    case Para::MutType::PerGenomeStepNoMinMaxBoundary:
      change_parameter_no_boundary(k, Para::mutation_step, random);
      break;
    case Para::MutType::PerGenomeStepWrapNormal:
      change_parameter_step_wrap_normal(k, Para::min_k, Para::max_k, Para::mutation_step, random);
      break;
    case Para::MutType::PerGenomeBinary:
      change_parameter_binary(k, Para::min_k, Para::max_k);
      break;
    case Para::MutType::PerGenomeMultiLocus:
      change_parameter_multi_locus(Para::min_k, Para::max_k, random);
      break;
    default:
      ParallelPrint::cerr("Molecule::mutate_Full() ERROR! Unkown value in Para::muttype\n");
      exit(-1);
      break;
    }
  }
}

double Molecule::get_k() const
{
  switch(Para::mut_type){
  case Para::MutType::PerGenomeStepWrap:
  case Para::MutType::PerGenomeStepNoMinBoundary:
  case Para::MutType::PerGenomeStepNoMinMaxBoundary:
  case Para::MutType::PerGenomeStepWrapNormal:
  case Para::MutType::PerGenomeBinary:
  case Para::MutType::PerGenomeMultiLocus:
    return k;
  case Para::MutType::PerGenomeStepNegative:
    return ((k>Para::min_k) ? k : Para::min_k);
    break;
  default:
    ParallelPrint::cerr("Molecule::get_k() ERROR! Unkown value in Para::muttype\n");
    exit(-1);
    break;
  }
}


void Molecule::change_parameter_step_wrap(double& parameter,const double min,const double max,const double step, RandomTBB2::reference random)
{
  parameter += step*random.FloatS();

  if(parameter<min){
    parameter = 2.*min - parameter;
  }else if(parameter>max){
    parameter = 2.*max - parameter;
  }
}

void Molecule::change_parameter_step_wrap_normal(double& parameter,const double min,const double max,const double step, RandomTBB2::reference random)
{
  parameter += random.Normal(0.0, step);

  if(parameter<min){
    parameter = 2.*min - parameter;
  }else if(parameter>max){
    parameter = 2.*max - parameter;
  }
}

void Molecule::change_parameter_negative(double& parameter,const double max,const double step, RandomTBB2::reference random)
{
  parameter += step*random.FloatS();

  if(parameter>max){
    parameter = 2.*max - parameter;
  }
}

void Molecule::change_parameter_no_boundary(double& parameter,const double step, RandomTBB2::reference random)
{
  parameter += step*random.FloatS();
}

void Molecule::change_parameter_binary(double& parameter,const double min,const double max)
{
  if(parameter>min){
    parameter = min;
  }
  else{
    parameter = max;
  }
}

void Molecule::change_parameter_multi_locus(const double min, const double max, RandomTBB2::reference random)
{
  /* Determine where the mutation occurs */
  std::size_t pos = random.Integer(genome.size());
  genome[pos] = !genome[pos];

  /* Update the k value */
  k = genome2k();

}

double Molecule::get_fitness() const
{
  return fitness;
}

void Molecule::set_fitness(double _fitness)
{
  fitness = _fitness;
}

void Molecule::replicate(const Molecule& rhs, RandomTBB2::reference random)
{
  *this = rhs;
  mutate(random);
}

std::ostream& operator<<(std::ostream& os, const Molecule& mol)
{
  os << ' ' << mol.get_k();
  os << ' ' << mol.get_fitness();
  os << ' ' << mol.get_cell_index();
  return os;
}

double Molecule::genome2k() const
{
  long n_loci_plus = 0;
  for(const auto& it: genome){
    if(it){
      ++n_loci_plus;
    }
  }

  return (2*n_loci_plus - static_cast<long>(genome.size())) * Para::mutation_step;
}
