#include "observer.hpp"

void Stat::reset(std::size_t n_cell)
{
  global_w_k.reset();
  
  between_norm_w_k.reset();
  between_norm_w_dk2.reset();
  between_norm_w_dk3.reset();

  within_norm_w_k.reset();
  within_norm_w_dk2.reset();
  within_norm_w_dk3.reset();

  cell_stat.resize(n_cell);

  for(auto& el : cell_stat){
    el.reset();
  }

  cell_level_hist_k.reset();

  mol_level_variance_k = 0;
  mol_level_central_moment_3rd_k = 0;
  mol_level_central_moment_4th_k = 0;
}

void CellStat::reset()
{
  w_k.reset();
  norm_w_k.reset();
  norm_w_dk2.reset();
  norm_w_dk3.reset();
  hist_k.reset();
}

Observer::Observer(const System& _system)
  : system(_system),
    time_plot_file(Para::fname_plot.c_str(), std::ios_base::app)
{
  /* Check if file is opened and plot for time=0 */
  if(time_plot_file.is_open()){
    print_header_time_plot();
  }
  else{
    std::cerr << "Observer::Observer() Cannot open " << Para::fname_plot << "\n";
    std::exit(-1);
  }

  if(!Para::fname_etc.empty()){
    etc_file.open(Para::fname_etc.c_str(), std::ios_base::app);

    if(!etc_file.is_open()){
      std::cerr << "Observer::Observer() Cannot open " << Para::fname_etc << "\n";
      std::exit(-1);
    }
  }

  /* Open file for plotting cell statistics */
  if(Para::trace_dynamics_of_each_cell){
    cell_stat_file.open(Para::fname_cell_stat.c_str(), std::ios_base::app);

    /* Check if file is opened and cell_stat for time=0 */
    if(cell_stat_file){
      print_header_cell_stat();
    }
    else{
      std::cerr << "Observer::Observer() Cannot open " << Para::fname_cell_stat << "\n";
      std::exit(-1);
    }
  }
}

void Observer::observe_system(unsigned long step)
{
  if(step%Para::plot_interval==0){
    calc_stat(step);

    /* Print statistics */
    print_time_plot();

    if(Para::trace_dynamics_of_each_cell){
      if(step%Para::cell_stat_interval==0){
	print_cell_stat();
      }
    }
  }
}

void Observer::calc_stat(unsigned long step)
{

  const System::MolVec& molvec = *system.molvec_curr_ptr;
  const CellPop& cellpop = *system.cellpop_curr_ptr;
  /* Reset statistics */
  stat.reset(cellpop.get_n_alive_cells());
      
  /* Fitness and catalysis of Molecule to be used with CovMat */
  std::vector<double> w_k;

  /* Calculate statistics independent of the average k within a Cell */
  for(auto& mol: molvec){
    /* Aggregate trait values */
    w_k.clear();
    w_k.push_back(mol.get_fitness());
    w_k.push_back(mol.get_k());

    /* Record data for Stat (global) */
    stat.global_w_k.add_sample(w_k);

    /* Record data for CellStat */
    stat.cell_stat[mol.get_cell_index()].w_k.add_sample(w_k);
    
    /* Record more detailes for cell_stat */
    stat.cell_stat[mol.get_cell_index()].hist_k.add_datum(mol.get_k());

  }

  /* Calculate the global statistics */
  stat.global_w_k.calc();

  const double n_mol_inverse = (stat.global_w_k.sample_size()) ? 1.0/stat.global_w_k.sample_size() : 0.0;
  const double gloabl_ave_w_inverse = (stat.global_w_k.ave(TraitType::fitness)>0.0) ? 1.0/stat.global_w_k.ave(TraitType::fitness) : 0.0;
  const double global_ave_k = stat.global_w_k.ave(TraitType::catalysis);
  
  /* Fitness and catalysis of Molecule to be used with CovMat */
  std::vector<double> w_dk2, w_dk3;
  
  /* Calculate statistics for each Cell */
  for(auto& cell_stat : stat.cell_stat){
    /* Calculate the average of this Cell's Covmat */
    cell_stat.w_k.calc();
    
    /* Add the k statistics of this Cell to between_k */
    w_k.clear();
    w_k.push_back(cell_stat.w_k.ave(TraitType::fitness) * gloabl_ave_w_inverse);
    w_k.push_back(cell_stat.w_k.ave(TraitType::catalysis));
    stat.between_norm_w_k.add_sample(w_k, cell_stat.w_k.sample_size());

    /* Add the w statistics of this Cell to between_w */
    w_dk2.clear();
    w_dk2.push_back(cell_stat.w_k.ave(TraitType::fitness) * gloabl_ave_w_inverse);
    w_dk2.push_back(pow(cell_stat.w_k.ave(TraitType::catalysis) - global_ave_k, 2));
    stat.between_norm_w_dk2.add_sample(w_dk2, cell_stat.w_k.sample_size());

    /* Add the w statistics of this Cell to between_w */
    w_dk3.clear();
    w_dk3.push_back(cell_stat.w_k.ave(TraitType::fitness) * gloabl_ave_w_inverse);
    w_dk3.push_back(pow(cell_stat.w_k.ave(TraitType::catalysis) - global_ave_k, 3));
    stat.between_norm_w_dk3.add_sample(w_dk3, cell_stat.w_k.sample_size());

    /* Add the k statistics of this Cell to between_k */
    stat.cell_level_hist_k.add_datum(cell_stat.w_k.ave(TraitType::catalysis), cell_stat.w_k.sample_size());
  
    try{
      /* Calculate frequency histogram of this Cell */
      cell_stat.hist_k.calc();
      
    }
    catch(std::bad_alloc& ba){
      std::cerr << "ERROR: " << ba.what() << "\n";
      exit(1);
    }

    stat.mol_level_variance_k += cell_stat.hist_k.variance() * cell_stat.hist_k.sample_size() * n_mol_inverse;
    stat.mol_level_central_moment_3rd_k += cell_stat.hist_k.central_moment_3rd() * cell_stat.hist_k.sample_size() * n_mol_inverse;
    stat.mol_level_central_moment_4th_k += cell_stat.hist_k.central_moment_4th() * cell_stat.hist_k.sample_size() * n_mol_inverse;
  }

  /* Calculate between-cell statistics */
  stat.between_norm_w_k.calc();
  stat.between_norm_w_dk2.calc();
  stat.between_norm_w_dk3.calc();
  stat.cell_level_hist_k.calc();
  
  /* Calculate statistics dependent on the average k within a Cell */
  for(auto& mol: molvec){
    CellStat& cell_stat = stat.cell_stat[mol.get_cell_index()];
    const double within_ave_w_inverse = (cell_stat.w_k.ave(TraitType::fitness)>0.0) ? 1.0/cell_stat.w_k.ave(TraitType::fitness) : 0.0;
    const double within_ave_k = cell_stat.w_k.ave(TraitType::catalysis);
	
    /* Aggregate trait values */
    w_k.clear();
    w_k.push_back(mol.get_fitness() * within_ave_w_inverse);
    w_k.push_back(mol.get_k());

    /* Add sample */
    cell_stat.norm_w_k.add_sample(w_k);
    
    /* Aggregate trait values */
    w_dk2.clear();
    w_dk2.push_back(mol.get_fitness() * within_ave_w_inverse);
    w_dk2.push_back(pow(mol.get_k() - within_ave_k, 2));

    /* Aggregate trait values */
    w_dk3.clear();
    w_dk3.push_back(mol.get_fitness() * within_ave_w_inverse);
    w_dk3.push_back(pow(mol.get_k() - within_ave_k, 3));

    /* Add sample */
    cell_stat.norm_w_dk2.add_sample(w_dk2);
    cell_stat.norm_w_dk3.add_sample(w_dk3);
  }

  /* Calculate within-cell statistics */
  for(auto& cell_stat : stat.cell_stat){
    /* Calculate within-cell statistics */
    cell_stat.norm_w_k.calc();
    cell_stat.norm_w_dk2.calc();
    cell_stat.norm_w_dk3.calc();

    /* Take average of within-cel statistics */
    stat.within_norm_w_k.add_sample(cell_stat.norm_w_k);
    stat.within_norm_w_dk2.add_sample(cell_stat.norm_w_dk2);
    stat.within_norm_w_dk3.add_sample(cell_stat.norm_w_dk3);
  }

  /* Calculate averages etc. */
  stat.within_norm_w_k.calc();
  stat.within_norm_w_dk2.calc();
  stat.within_norm_w_dk3.calc();
}

void Observer::print_header_time_plot()
{
  std::ostream& os = time_plot_file;
  unsigned int counter = 1;

  os << "#t(" << counter++ << ")";

  os << ' ' << "V(" << counter++ << ")";

  os << ' ' << "k(" << counter++ << ")";

  os << ' ' << "mk2(" << counter++ << ")";
  os << ' ' << "mk3(" << counter++ << ")";
  os << ' ' << "mk4(" << counter++ << ")";
  os << ' ' << "mwk(" << counter++ << ")";
  os << ' ' << "mwk2(" << counter++ << ")";
  os << ' ' << "mwk3(" << counter++ << ")";

  os << ' ' << "ck2(" << counter++ << ")";
  os << ' ' << "ck3(" << counter++ << ")";
  os << ' ' << "ck4(" << counter++ << ")";
  os << ' ' << "cwk(" << counter++ << ")";
  os << ' ' << "cwk2(" << counter++ << ")";
  os << ' ' << "cwk3(" << counter++ << ")";

  os << std::endl;
}

void Observer::print_time_plot()
{
  std::ostream& os = time_plot_file;

  /* Print */
  os << system.get_n_generation();

  os << ' ' << system.cellpop_curr_ptr->get_n_alive_cells();

  os.precision(6);
  os.setf(std::ios::scientific);

  os << ' ' << stat.global_w_k.ave(TraitType::catalysis);

  os << ' ' << stat.mol_level_variance_k;
  os << ' ' << stat.mol_level_central_moment_3rd_k;
  os << ' ' << stat.mol_level_central_moment_4th_k;
  os << ' ' << stat.within_norm_w_k.cov(TraitType::fitness, TraitType::catalysis);
  os << ' ' << stat.within_norm_w_dk2.cov(TraitType::fitness, TraitType::catalysis);
  os << ' ' << stat.within_norm_w_dk3.cov(TraitType::fitness, TraitType::catalysis);
  
  
  os << ' ' << stat.cell_level_hist_k.variance();
  os << ' ' << stat.cell_level_hist_k.central_moment_3rd();
  os << ' ' << stat.cell_level_hist_k.central_moment_4th();
  os << ' ' << stat.between_norm_w_k.cov(TraitType::fitness, TraitType::catalysis);
  os << ' ' << stat.between_norm_w_dk2.cov(TraitType::fitness, TraitType::catalysis);
  os << ' ' << stat.between_norm_w_dk3.cov(TraitType::fitness, TraitType::catalysis);
  
  os.precision(6);
  os.unsetf(std::ios::scientific);

  os << std::endl;
}

bool Observer::extinct() const
{
  return false;
}

void Observer::print_header_cell_stat()
{
  std::ostream& os = cell_stat_file;
  unsigned int counter = 1;

  os << "#t(" << counter++ << ")";

  os << ' ' << "kall(" << counter++ << ")";

  os << ' ' << "ID(" << counter++ << ")";

  os << ' ' << "pID(" << counter++ << ")";

  os << ' ' << "V(" << counter++ << ")";

  /* Moments */
  os << ' ' << "k(" << counter++ << ")";
  os << ' ' << "mk2(" << counter++ << ")";
  os << ' ' << "mk3(" << counter++ << ")";
  os << ' ' << "mk4(" << counter++ << ")";
  os << ' ' << "mwk(" << counter++ << ")";
  os << ' ' << "mwk2(" << counter++ << ")";
  os << ' ' << "mwk3(" << counter++ << ")";
  
  // os << ' ' << "skew(" << counter++ << ")";
  // os << ' ' << "kurt(" << counter++ << ")";

  /* Frequencies */
  os << ' ' << "min_bin_ind(" << counter++ << ")";
  os << ' ' << "bin_width(" << counter++ << ")";

  os << ' ' << "bin1(" << counter++ << ")";
  os << ' ' << "bin2(" << counter++ << ")";
  os << ' ' << "...";

  os << std::endl;
}

void Observer::print_cell_stat()
{
  std::ostream& os = cell_stat_file;
  const CellPop& cellpop = *system.cellpop_curr_ptr;
  
  for(std::size_t ind=0, end=cellpop.get_n_alive_cells(); ind!=end; ++ind){
    auto& cell = stat.cell_stat[ind];
    
    os << system.get_n_generation();

    os.setf(std::ios::scientific);

    os.precision(6);
    
    os << ' ' << stat.global_w_k.ave(TraitType::catalysis);

    /* Cell statistics */
    
    os << ' ' << ind;
    os << ' ' << cellpop.get_parent_index(ind);
    os << ' ' << cell.hist_k.sample_size();

    /* Moments */
    os << ' ' << cell.hist_k.mean();

    os << ' ' << cell.hist_k.variance();
    os << ' ' << cell.hist_k.central_moment_3rd();
    os << ' ' << cell.hist_k.central_moment_4th();
    os << ' ' << cell.norm_w_k.cov(TraitType::fitness, TraitType::catalysis);;
    os << ' ' << cell.norm_w_dk2.cov(TraitType::fitness, TraitType::catalysis);;
    os << ' ' << cell.norm_w_dk3.cov(TraitType::fitness, TraitType::catalysis);;


    /* Frequencies */
    os << ' ' << cell.hist_k.index_of_min_bin();
    os << ' ' << cell.hist_k.bin_width();

    for(auto& freq : cell.hist_k.frequencies()){
      os << ' ' << freq;
    }

    os << std::endl;
    os.precision(6);
    os.unsetf(std::ios::scientific);
  }

}

// bool Observer::system_is_dead() const
// {
//   return (stat.f < 1e-6);
// }

void Observer::print_one_cell()
{
  // if(!etc_file.is_open()){
  //   std::cerr << "Observer::print_one_cell() ERROR: file not open\n";
  //   exit(1);
  // }

  // const CellPop::CellVec& cellvec = system.cellpop_curr_ptr->cellvec;

  // unsigned long ind = cellvec[0].molecules_ptr[0]->get_cell_index();
  // CellStat& cstat = stat.cell_stat[ind];

  // std::ostream& os = etc_file;
  // os.precision(10);
  // os.setf(std::ios::scientific);

  // for(auto iter=cellvec[0].molecules_ptr.begin(), end=cellvec[0].molecules_ptr.end(); iter!=end; ++iter){
  //   os << system.get_n_generation();
  //   os << ' ' << cellvec[0].molecules_ptr.size() - static_cast<unsigned long>(end - iter);;
  //   os << ' ' << cstat.n_mol;
  //   os << ' ' << cstat.k.ave(TraitType::k);
  //   os << ' ' << cstat.w.ave(FitnessType::w);
  //   os << ' ' << (*iter)->k;
  //   os << ' ' << (*iter)->get_fitness();
  //   os << std::endl;
  // }

  // os.precision(6);
  // os.unsetf(std::ios::scientific);
}

void Observer::print_best_mol()
{
  // if(!etc_file.is_open()){
  //   std::cerr << "Observer::print_best_mol() ERROR: file not open\n";
  //   exit(1);
  // }

  // const System::MolVec& molvec = *system.molvec_curr_ptr;
  // double fitness = -1 * std::numeric_limits<double>::infinity();
  // const Molecule* best_mol = 0;

  // /* Calculate statistics */
  // for(auto iter=molvec.begin(), end=molvec.end(); iter!=end; ++iter){
  //   const Molecule& mol = *iter;

  //   if(fitness < mol.get_fitness()){
  //     fitness = mol.get_fitness();
  //     best_mol = &mol;
  //   }
  // }

  // if(best_mol == 0){
  //   return;
  // }

  // //const CellPop::CellVec& cellvec = system.cellpop_curr_ptr->cellvec;
  // //const Cell& cell = cellvec[best_mol->get_cell_index()];

  // std::ostream& os = etc_file;

  // // os.precision(3);
  // // os.setf(std::ios::scientific);
    
  // // for(auto iter=cell.molecules_ptr.begin(), end=cell.molecules_ptr.end(); iter!=end; ++iter){
  // //   const Molecule& mol = **iter;
  // //   os << system.get_n_generation() << ' ' << mol << std::endl;
  // // }

  // // os.precision(6);
  // // os.unsetf(std::ios::scientific);


  // os << system.get_n_generation();

  // os << ' ' << stat.cell_stat[best_mol->get_cell_index()].n_mol;

  // os.precision(3);
  // os.setf(std::ios::scientific);
  // os << ' ' << stat.cell_stat[best_mol->get_cell_index()].w.ave(FitnessType::w);
  // os << ' ' << stat.cell_stat[best_mol->get_cell_index()].k.ave(TraitType::k);

  // os << ' ' << *best_mol;

  // os.precision(6);
  // os.unsetf(std::ios::scientific);

  // os << std::endl;
}

void Observer::print_all_cells()
{
  // if(!etc_file.is_open()){
  //   std::cerr << "Observer::print_all_cells() ERROR: file not open\n";
  //   exit(1);
  // }

  // std::ostream& os = etc_file;
  // os.precision(10);
  // os.setf(std::ios::scientific);

  // for(auto iter=stat.cell_stat.begin(), end=stat.cell_stat.end(); iter!=end; ++iter){
  //   CellStat& cstat = *iter;

  //   //const double p = static_cast<double>(cstat.nP)/(cstat.nP + cstat.nQ);
  //   //const double q = static_cast<double>(cstat.nQ)/(cstat.nP + cstat.nQ);


  //   os << system.get_n_generation();
  //   os << ' ' << stat.cell_stat.size() - static_cast<unsigned long>(end - iter);
  //   os << ' ' << cstat.n_mol;
  //   os << ' ' << cstat.w.ave(FitnessType::w);
  //   os << ' ' << cstat.k.ave(TraitType::k);
  //   os << std::endl;
  // }

  // os.precision(6);
  // os.unsetf(std::ios::scientific);
  
}

bool Observer::should_stop()
{
  const double k = stat.global_w_k.ave(TraitType::catalysis);

    switch(Para::mut_type){
    case Para::MutType::PerGenomeStepWrap:
    case Para::MutType::PerGenomeStepNegative:
    case Para::MutType::PerGenomeStepNoMinBoundary:
    case Para::MutType::PerGenomeStepNoMinMaxBoundary:
    case Para::MutType::PerGenomeStepWrapNormal:
      if(k<Para::min_k + 5*sqrt(Para::mutation_step) || Para::max_k - 5*sqrt(Para::mutation_step) < k){
	std::cerr << "Observer::should_stop() Stop condition is met\n";
	return true;
      }
      else{
	return false;
      }
    case Para::MutType::PerGenomeBinary:
    case Para::MutType::PerGenomeMultiLocus:
      return false;      
    default:
      ParallelPrint::cerr("Observer::should_stop() ERROR! Unkown value in Para::muttype\n");
      exit(-1);
      break;
    }

  
}
