// Standard Library
#include <iostream>
#include <vector>
#include <algorithm>
#include <limits>

// External Libraries

// My Libraries
#include "assert.hpp"

// Internal 
#include "para.hpp"
#include "covmat.hpp"
#include "system.hpp"
#include "histogram.hpp"

#ifndef OBSERVER
#define OBSERVER

namespace TraitType{
  enum Type {fitness, catalysis};
}

/* Statistics of Cells */
struct CellStat {
  void reset();

  /* Covariance between w_ij and k_ij of Molecules within a Cell */
  CovMatDirect<2> w_k;
  
  /* Covariance between w_ij/w_i and k_ij of Molecules within a Cell,
     where w_i is within-cell average */
  CovMatDirect<2> norm_w_k;

  /* Covariance between w_ij/w_i and (k_ij - k_i)^2 of Molecules within a
     Cell, where w_i is within-cell average */
  CovMatDirect<2> norm_w_dk2;

    /* Covariance between w_ij/w_i and (k_ij - k_i)^3 of Molecules within a
     Cell, where w_i is within-cell average */
  CovMatDirect<2> norm_w_dk3;

  /* Individal k values of all Molecules. Used to calculate kurtosis and
     skewness */
  Histogram hist_k;
};

struct Stat {
  void reset(std::size_t n_cell);

  /* Covariance between w_ij and k_ij */
  CovMatDirect<2> global_w_k;

  /* Covariance between w_i/w and k_i of Cells, where where w is global
     average */
  CovMatDirect<2> between_norm_w_k;

  /* Covariance between w_i/w and (k_i-k)^2 of Cells, where where w is
     global average */
  CovMatDirect<2> between_norm_w_dk2;

  /* Covariance between w_i/w and (k_i-k)^3 of Cells, where where w is
     global average */
  CovMatDirect<2> between_norm_w_dk3;

  /* Average covariance between w_ij/w_i and k_ij */
  CovMatAverage<2> within_norm_w_k;

  /* Average covariance between w_ij/w_i and (k_ij-k_i)^2 */
  CovMatAverage<2> within_norm_w_dk2;

  /* average covariance between w_ij/w_i and (k_ij-k_i)^3 */
  CovMatAverage<2> within_norm_w_dk3;

  /* Statistics of Molecules in each Cell */
  std::vector<CellStat> cell_stat;

  /* Histogram of k_i. */
  Histogram cell_level_hist_k;

  /* Average central moments of k_ij. */
  double mol_level_variance_k;
  double mol_level_central_moment_3rd_k;
  double mol_level_central_moment_4th_k;
  
};

class Observer {
public:
  Observer(const System& _system);
  void observe_system(unsigned long step);
  bool should_stop();

private:
  void calc_stat(unsigned long step);
  void print_header_time_plot();
  void print_time_plot();
  bool extinct() const;
  void print_header_cell_stat();
  void print_cell_stat();
  void print_one_cell();
  void print_all_cells();
  void print_best_mol();

  const System& system;
  Stat stat;
  std::ofstream time_plot_file;
  std::ofstream cell_stat_file;
  std::ofstream etc_file;
};

#endif
