#include "random_wrapper2.hpp"

/* Definition of static data members follows: */

/* This will be zeroed by default. */
tbb::atomic<unsigned> RandomWrapper2::counter;

/* This will be empty by default */
std::vector<std::seed_seq::result_type> RandomWrapper2::seeds;

/* This is used to allow each thread to send a whole chunk of texts to
   output stream. */
RandomWrapper2::StreamMutex RandomWrapper2::stream_mutex;

/* Grobal variable(s) used by users of this module. */
RandomTBB2 random_tbb_2;

RandomWrapper2::RandomWrapper2()
  : index(counter++),
    boolean(0.5),
    uniform_real(0.0, 1.0),
    uniform_real_N(0.0, std::nextafter(1.0,2.0)),
    uniform_real_W(-1.0, std::nextafter(1.0,2.0)),
    uniform_real_S(std::nextafter(-0.5, 0.), 0.5),
    normal()
{
  /* Error check. Before constructing "RandomWrapper2", "seeds" must be
     set. To this end, use "set_seed()". */  
  if(seeds.empty()){
    {
      StreamMutex::scoped_lock lock(stream_mutex);
      std::cerr << "RandomWrapper2::RandomWrapper2() Error. RandomWrapper2::seeds is uninitialized."  << std::endl;
    }
    exit(-1);
  }

  /* Error check. The number of RNG cannot exceeds max_n_gen */
  if(index >= max_n_gen()){
    {
      StreamMutex::scoped_lock lock(stream_mutex);
      std::cerr << "RandomWrapper2::RandomWrapper2() Error. The number of random number generators generated exceeded max_n_gen."  << std::endl;
    }
    exit(-1);
  }

  /* Seed this RNG. */
  seed(seeds[index]);

  /* Report */
  {
    StreamMutex::scoped_lock lock(stream_mutex);
    std::cerr << "RandomWrapper2::RandomWrapper2() Default constructed. Seed = " << seeds[index] << std::endl;
  }

}

void RandomWrapper2::set_seed(unsigned long seed, unsigned n_gen)
{
  /* Error check on "n_gen". */
  if(n_gen == 0){
    {
      StreamMutex::scoped_lock lock(stream_mutex);
      std::cerr << "RandomWrapper2::set_seed() Error. n_gen must be greater than zero."  << std::endl;
    }
    exit(-1);
  }

  /* Error check on re-seeding. */
  if(max_n_gen()){
    {
      StreamMutex::scoped_lock lock(stream_mutex);
      std::cerr << "RandomWrapper2::set_seed() Error. Re-seeding is prohibited."  << std::endl;
    }
    exit(-1);
  }

  /* Set the number of seeds. */
  seeds.resize(n_gen);

  /* Generates seeds */
  std::seed_seq seq({seed});
  seq.generate(seeds.begin(), seeds.end());
}

unsigned RandomWrapper2::max_n_gen()
{
  return seeds.size();
}

int RandomWrapper2::Integer(int N)
{
  return uniform_int(*this, decltype(uniform_int)::param_type(0,N-1));
}

unsigned RandomWrapper2::Integer(unsigned N)
{
  return uniform_unsigned(*this, decltype(uniform_unsigned)::param_type(0,N-1));
}

long RandomWrapper2::Integer(long N)
{
  return uniform_long(*this, decltype(uniform_long)::param_type(0,N-1));
}

unsigned long RandomWrapper2::Integer(unsigned long N)
{
  return uniform_unsigned_long(*this, decltype(uniform_unsigned_long)::param_type(0,N-1));
}

// int RandomWrapper2::operator()(int N)
// {
//   return Integer(N);
// }

// unsigned RandomWrapper2::operator()(unsigned N)
// {
//   return Integer(N);
// }

// long RandomWrapper2::operator()(long N)
// {
//   return Integer(N);
// }

// unsigned long RandomWrapper2::operator()(unsigned long N)
// {
//   return Integer(N);
// }

bool RandomWrapper2::Prob(double p)
{
  return bernoulli(*this, std::bernoulli_distribution::param_type(p));
}

bool RandomWrapper2::Boolean()
{
  return boolean(*this);
}

double RandomWrapper2::Float()
{
  return uniform_real(*this);
}

double RandomWrapper2::FloatN()
{
  return uniform_real_N(*this);
}

double RandomWrapper2::FloatW()
{
  return uniform_real_W(*this);
}

double RandomWrapper2::FloatS()
{
  return uniform_real_S(*this);
}

int RandomWrapper2::Binomial(int n_trial, double p)
{
  return binomial_int(*this, decltype(binomial_int)::param_type(n_trial, p));
}

unsigned RandomWrapper2::Binomial(unsigned n_trial, double p)
{
  return binomial_unsigned(*this, decltype(binomial_unsigned)::param_type(n_trial, p));
}

long RandomWrapper2::Binomial(long n_trial, double p)
{
  return binomial_long(*this, decltype(binomial_long)::param_type(n_trial, p));
}

unsigned long RandomWrapper2::Binomial(unsigned long n_trial, double p)
{
  return binomial_unsigned_long(*this, decltype(binomial_unsigned_long)::param_type(n_trial, p));
}

double RandomWrapper2::Normal(double mean, double stddev)
{
  return normal(*this, decltype(normal)::param_type(mean, stddev));
}
