/*
  PURPOSE

  I want to use <random> with Intel Threading Building Blocks (tbb). For
  the sake of concurrency, each thread must have a distinct "Random"
  object initialized with a distinct seed. To this end, I use
  "tbb::enumerable_thread_specific", which allows each thread to have
  its local storage that is not shared by the other threads. To use
  "tbb::enumerable_thread_specific", I created a wrapper (called
  RandomWrapper2) of a random number generator from <random>.

  ------------------------------------------------------------------
  USAGE

  The code defines a global variable "random_tbb_2", which holds an
  instance of "RandomTBB2" (which is a specialization of
  "tbb::enumerable_thread_specific" with "RandomWrapper" as a template
  parameter). "random_tbb_2" alows one to use a thread-specific "Random"
  object. Before using "Random" objects, one must first set a seed. To
  do so, use:

  RandomWrapper2::set_seed(12345UL);

  One can now use "random_tbb_2". To access a thread-specific "Random"
  object, do

  RandomTBB2::reference X = random_tbb_2.local();

  Now, "X" is a thread-specific "Random" object. For example, one can do
  
  row = static_cast<unsigned>(X.Real()*nrow + 1);

  ------------------------------------------------------------------
  IMPLEMENTATION

  Apparently, "enumerable_thread_specific<T>" construct "T" objects only
  when it is needed to do so (e.g., by requested by a call to
  "local()"). Each thread will get one "T" object, which behaves, so to
  speak, as a "static" data of a thread. "enumerable_thread_specific<T>"
  constructs "T" objects with the default constructor (unless an
  exemplar is supplied). Thus, a proper definition of the default
  constructer of "T" is essential.

  The default constructor of "RandomWrapper" will construct a "Random"
  object (by the default constructer) and re-seed it with a seed unique
  to each "Random" object. The seeds are generated from two numbers
  (unsigned long) : a base, which is common to all "RandomWrapper"
  objects (see "set_seed()" below) and numbers unique to each
  "RandomWrapper" object. These two numbers are stored in
  "std::vector<unsigned long>", which is, in turn, given to
  "Random::Reseed()" to re-seed "Random" objects. 

  ------------------------------------------------------------------
  LICENSE

  This code uses a code taken from The GNU Scientific Library
  (GSL). Since GSL is distributed under The GNU General Public License
  (GPL), this code, if distributed, is also under GPL.
*/

#include <iostream>
#include <vector>
#include <unistd.h>
#include <random>
#include <cmath>

#include "tbb/tbb.h"

#ifndef RANDOM_WRAPPER_2
#define RANDOM_WRAPPER_2

class RandomWrapper2 : public std::mt19937_64 {
public:
  RandomWrapper2();

  /* Calling "set_seed()" is required before using this class. Calling
     it more than once is prohibited. In other words, re-seeding is not
     allowed. "n_gen" is the max number of random number generators that
     can be generated (the number of seeds to be prepared). */
  static void set_seed(unsigned long seed, unsigned n_gen);

  /* Maximum number of RNGs that can be generated. It is the length of
     "seeds". */
  static unsigned max_n_gen();

  /* Defining a function named f in a derived class (Derived) hides all
     members named f in its base classe (Base). In order to use Base::f
     while Derived::f is defined, you have to use the "using"
     declaration. */
  using std::mt19937_64::operator();

  /* Interger-type random variable. The naming convention of the
     following functions follows RandomLib by C.F.F. Karney. */
  
  /* Returns a non-negative random integer smaller than N  */
  int Integer(int N);
  unsigned Integer(unsigned N);
  long Integer(long N);
  unsigned long Integer(unsigned long N);

  template<class IntType>
  IntType operator()(IntType N);

  /* It returns true with a probability p */
  bool Prob(double p);

  /* It returns true or false with an equal probablity */
  bool Boolean();

  /* Returns a number sampled from a binomial distribution */
  int Binomial(int n_trial, double p);
  unsigned Binomial(unsigned n_trial, double p);
  long Binomial(long n_trial, double p);
  unsigned long Binomial(unsigned long n_trial, double p);

  /* Returns a set of numbers sampled from a multinomial distribution
     formed by N trials from an underlying distribution
     p[0]...p[K-1].. The result is stored in n[]. */
  template<class IntType>
  void Multinomial(const std::size_t K, const IntType N, const double p[], IntType n[]);
  
  /* Float-type random variable. The naming convention of the following
     functions follows RandomLib by C.F.F. Karney. */

  /* It returns a random number in [0,1). */
  double Float();

  /* It returns a random double in [0,1]. */
  double FloatN();

  /* It returns a random double in [-1,1]. */
  double FloatW();

  /* It returns a random double in (-0.5,0.5). */
  double FloatS();

  /* It returns a number sampled from a Normal distribution with mean
     and stddev. */
  double Normal(double mean, double stddev);

  
private:
  /* Forbid coping so users don't accidentally copy Random. */
  RandomWrapper2(const RandomWrapper2&) = delete;
  RandomWrapper2& operator=(const RandomWrapper2&) = delete;

  /* "counter" counts the number of RNGs generated  */
  static tbb::atomic<unsigned> counter;

  /* Seeds used to seed RNGs */
  static std::vector<std::seed_seq::result_type> seeds;

  /* Mutex to resolve race with output */
  typedef tbb::spin_mutex StreamMutex;
  static StreamMutex stream_mutex;

  /* A unique index of each instance. It runs 0, 1, ... */
  const unsigned index;

  /* The followings are predefined set of distributions. */
  std::uniform_int_distribution<int> uniform_int;
  std::uniform_int_distribution<unsigned> uniform_unsigned;
  std::uniform_int_distribution<long> uniform_long;
  std::uniform_int_distribution<unsigned long> uniform_unsigned_long;

  std::bernoulli_distribution bernoulli;
  std::bernoulli_distribution boolean;

  std::binomial_distribution<int> binomial_int;
  std::binomial_distribution<unsigned> binomial_unsigned;
  std::binomial_distribution<long> binomial_long;
  std::binomial_distribution<unsigned long> binomial_unsigned_long;

  std::uniform_real_distribution<double> uniform_real;
  std::uniform_real_distribution<double> uniform_real_N;
  std::uniform_real_distribution<double> uniform_real_W;
  std::uniform_real_distribution<double> uniform_real_S;

  std::normal_distribution<> normal;
};

typedef tbb::enumerable_thread_specific<RandomWrapper2> RandomTBB2;

/* "random_tbb_2" is what users should directy access. */
extern RandomTBB2 random_tbb_2;

/* Returns a non-negative random integer smaller than N. IntType is
   deduced by template argument deduction. */
template<class IntType>
IntType RandomWrapper2::operator()(IntType N)
{
  return Integer(N);
}

/* This function computes a random sample n[K] from the multinomial
 * distribution formed by N trials from an underlying distribution
 * p[K]. The distribution function for n[K] is,
 *
 *      P(n_1, n_2, ..., n_K) =
 *           (N!/(n_1! n_2! ... n_K!)) p_1^n_1 p_2^n_2 ... p_K^n_K
 *
 * where (n_1, n_2, ..., n_K) are nonnegative integers with sum_{k=1}^K
 * n_k = N, and (p_1, p_2, ..., p_K) is a probability distribution with
 * \sum p_i = 1. If the array p[K] is not normalized then its entries
 * will be treated as weights and normalized appropriately. The arrays
 * n[] and p[] must both be of length K.
 *
 * Random variates are generated using the conditional binomial method
 * (see C.S. Davis, The computer generation of multinomial random
 * variates, Comp. Stat. Data Anal. 16 (1993) 205217 for details).
 *
 * The code is taken from The GNU Scientific Library. Thus, this code, if
 * distributed, is subject to The GNU General Public License (GPL).
 *
 */
template<class IntType>
void RandomWrapper2::Multinomial(const std::size_t K, const IntType N, const double p[], IntType n[])
{
  /* p[k] may contain non-negative weights that do not sum to 1.0.
   * Even a probability distribution will not exactly sum to 1.0
   * due to rounding errors. 
   */
  double norm = 0.0;
  for (std::size_t k = 0; k < K; k++)
    {
      norm += p[k];
    }

  double sum_p = 0.0;
  IntType sum_n = 0;
  for (std::size_t k = 0; k < K; k++)
    {
      if (p[k] > 0.0)
	{
	  //n[k] = gsl_ran_binomial (ran, p[k] / (norm - sum_p), N - sum_n);
	  n[k] = Binomial(N - sum_n, p[k] / (norm - sum_p));
	}
      else
	{
	  n[k] = 0;
	}

      sum_p += p[k];
      sum_n += n[k];
    }
}

#endif
