#include "system.hpp"

System::System():
  nomultithread(Para::n_cpu == 1),
  molvec_0(Para::n_molecule),
  molvec_1(Para::n_molecule),
  molvec_curr_ptr(&molvec_0),
  molvec_next_ptr(&molvec_1),
  cellpop_curr_ptr(&cellpop_0),
  cellpop_next_ptr(&cellpop_1),
  time(0),
  fitness_sum(0.0),
  fitness_vec(Para::n_molecule)
{
  init();
}

unsigned long System::get_n_generation() const
{
  return time;
}

void System::swap_generation()
{
  std::swap(molvec_curr_ptr, molvec_next_ptr);
  std::swap(cellpop_curr_ptr, cellpop_next_ptr);
}

void System::init()
{
  MolVec& molvec_curr = *molvec_curr_ptr;
  CellPop& cellpop_curr = *cellpop_curr_ptr;

  /* Calculate how many cells are needed */
  std::size_t n_cells = Para::n_molecule/Para::division_threshold;
  if(Para::n_molecule%Para::division_threshold){
    ++n_cells;
  }

  /* Make Cells as much as needed */
  cellpop_curr.reset(n_cells);

  std::size_t cell_index = 0;

  /* Initialize Molecules */
  for(unsigned long i=0;i<Para::n_molecule;++i){
    /* If the size of the current Cell has reached the limit */
    if(cellpop_curr.cell_size(cell_index) == Para::division_threshold){
      /* Change to a new Cell */
      ++cell_index;
    }

    /* Add this Molecule to a Cell */
    cellpop_curr.add_mol(cell_index, &(molvec_curr[i]));

    /* Remember which Cell this Molecule belongs to */
    molvec_curr[i].set_cell_index(cell_index);
  }
  
  calc_fitness();
}

void System::calc_fitness()
{
  fitness_sum = cellpop_curr_ptr->calc_fitness();
}

void System::run_one_step()
{
  fisher_wright();
  divide_cell();
  swap_generation();
  calc_fitness();

  /* Update the time */
  ++time;

  /* Error check */
  try{
    Assert<GeneralError>(!ASSERT::ERROR_CHECK|| check_consistency());
  }
  catch(GeneralError){
    ParallelPrint::cerr("System::run_one_step() ERROR: The model's state is inconsistent\n");
    exit(-1);
  }
}

void System::fisher_wright()
{
  /*debug*/
  //fisher_wright_parallel();
  
  if(nomultithread){
    fisher_wright_serial();
  }
  else{
    fisher_wright_parallel();
  }
}

class System::SetFitnessVector {
  const MolVec& molvec_curr;
  std::vector<double>& fitness_vec;
    
public:
  double my_sum_fitness;
  void operator()(const tbb::blocked_range<std::size_t>& range) {

    /* Use a local to speed up. We set the value of the local to that of
       the member because this instance of the class might have been
       already called, in which case its member contains the result of
       the previous call, which should not be discarded (this is
       explained in TBB User Guide). */
    double sum_fitness = my_sum_fitness;

    /* Set fitness vector and sum up fitness */
    for(std::size_t ind=range.begin(), end=range.end(); ind!=end; ++ind){
      fitness_vec[ind] = molvec_curr[ind].get_fitness();
      sum_fitness += fitness_vec[ind];
    }

    my_sum_fitness = sum_fitness;
  }
  
  /* Constructor */
  SetFitnessVector(const MolVec& _molvec_curr, std::vector<double>& _fitness_vec) :
    molvec_curr(_molvec_curr),
    fitness_vec(_fitness_vec),
    my_sum_fitness(0.0)
  {}
  
    /* Split constructor, which is used for parallel_reduce */
  SetFitnessVector(SetFitnessVector& x, tbb::split) :
    molvec_curr(x.molvec_curr),
    fitness_vec(x.fitness_vec),
    my_sum_fitness(0.0)
  {}

  /* A function used for parallel_reduce */
  void join(const SetFitnessVector& y){my_sum_fitness += y.my_sum_fitness;}
};

class System::DoFisherWright {
  const MolVec& molvec_curr;
  MolVec& molvec_next;
  CellPop& cellpop_curr;
  const std::vector<double>& fitness_vec;
    
public:
  void operator()(const tbb::blocked_range<std::size_t>& range) const {

    /* Fisher-Wright process */
    std::size_t curr_index;
    RandomTBB2::reference random = random_tbb_2.local();
    std::discrete_distribution<> select(fitness_vec.begin(), fitness_vec.end());
    for(std::size_t next_index=range.begin(), end=range.end(); next_index != end; ++next_index){
      /* Choose a Molecule with a probability proportional to its fitness */
      curr_index = select(random);

      /* Replicate a Molecule. Note that the cell_index of the
	 next-generation Molecule is set to the same value as that of
	 its parent Molecule at this point. cell_index will be updated
	 by divide_cell() later. */
      molvec_next[next_index].replicate(molvec_curr[curr_index], random);

      /* Record the parent of the new Molecule */
      molvec_next[next_index].set_parent_index(curr_index);

      /* Record the next-gen size of the Cell. add_nextgen_size() uses
	 tbb::atomic and is thus thread-safe. */
      cellpop_curr.add_nextgen_size(molvec_curr[curr_index].get_cell_index());
    }
  }
  DoFisherWright(const MolVec& _molvec_curr, MolVec& _molvec_next, CellPop& _cellpop_curr, const std::vector<double>& _fitness_vec) :
    molvec_curr(_molvec_curr),
    molvec_next(_molvec_next),
    cellpop_curr(_cellpop_curr),
    fitness_vec(_fitness_vec)
  {}
};

void System::fisher_wright_parallel()
{
  const MolVec& molvec_curr = *molvec_curr_ptr;

  /* Set a fitness vector */
  SetFitnessVector set_fitness_vector(molvec_curr,fitness_vec);
  tbb::parallel_reduce(tbb::blocked_range<std::size_t>(0, molvec_curr.size()), set_fitness_vector, tbb::static_partitioner());

  /* If sum_fitness==0, we have to re-set a fitness vector because it
     will cause a trouble to std::discrete_distribution */
  if(set_fitness_vector.my_sum_fitness==0.0){
    std::fill(fitness_vec.begin(),fitness_vec.end(), 1.0);
  }

  /* Fisher-Wright process */
  tbb::parallel_for(tbb::blocked_range<std::size_t>(0, molvec_curr.size()), DoFisherWright(molvec_curr,*molvec_next_ptr,*cellpop_curr_ptr,fitness_vec), tbb::static_partitioner());
}

void System::fisher_wright_serial()
{
  const MolVec& molvec_curr = *molvec_curr_ptr;
  MolVec& molvec_next = *molvec_next_ptr;
  CellPop& cellpop_curr = *cellpop_curr_ptr;

  /* Set a fitness vector */
  double sum_fitness = 0.0;
  for(std::size_t ind=0, end=molvec_curr.size(); ind<end; ++ind){
    fitness_vec[ind] = molvec_curr[ind].get_fitness();
    sum_fitness += fitness_vec[ind];
  }

  /* If sum_fitness==0, we have to re-set a fitness vector because it
     will cause a trouble to std::discrete_distribution */
  if(sum_fitness==0.0){
    std::fill(fitness_vec.begin(),fitness_vec.end(), 1.0);
  }

  /* Prepare RandomSelect to select Molecules */
  std::discrete_distribution<> select(fitness_vec.begin(), fitness_vec.end());
  RandomTBB2::reference random = random_tbb_2.local();

  /* Fisher-Wright process */
  std::size_t index;
  for(auto iter=molvec_next.begin(), end=molvec_next.end(); iter!=end; ++iter){
    /* Choose a Molecule with a probability proportional to its fitness */
    index = select(random);

    /* Replicate a Molecule. Note that Molecule of the next generation
       also copies cell_index of the Molecule of the current
       generation. cel_index of Molecules of the next generation needs
       to be updated. This will be done by System::divide_cell(). */
    iter->replicate(molvec_curr[index], random);
    
    /* Record the parent of the new Molecule */
    iter->set_parent_index(index);

    /* Record the next-gen size of the Cell */
    cellpop_curr.add_nextgen_size(molvec_curr[index].get_cell_index());
  }
}

/* Update the cell_index of next-generation Molecules. */
void System::divide_cell()
{
  /*debug*/
  //divide_cell_parallel();
  
  if(nomultithread){
    divide_cell_serial();
  }
  else{
    divide_cell_parallel();
  }
}

void System::divide_cell_serial()
{
  MolVec& molvec_next = *molvec_next_ptr;  
  CellPop& cellpop_curr = *cellpop_curr_ptr;
  CellPop& cellpop_next = *cellpop_next_ptr;

  /* Set the daughter_index of current-generation Cells */
  std::size_t n_nextgen_cells = cellpop_curr.set_daughter_cell_index();

  /* Make next-generation Cells */
  cellpop_next.reset(n_nextgen_cells);

  /* Assign next-generation Molecules to next-generation Cells */
  RandomTBB2::reference random = random_tbb_2.local();
  std::size_t curr_cell_index, next_cell_index;
  for(auto iter=molvec_next.begin(), end=molvec_next.end(); iter!=end; ++iter){
    /* Get the index of the current-generation Cell whose Molecule gave
       rise to this Molecule in the next generation */
    curr_cell_index = iter->get_cell_index();

    /* Choose which daughter Cell this Molecule belongs to */
    const std::vector<std::size_t>& daughter_index = cellpop_curr.get_daughter_index(curr_cell_index);

    /* No cell division */
    if(daughter_index.size() == 1){
      next_cell_index = daughter_index[0];
    }
    /* Cell division occurs */
    else{
      /* Choose randomly a daughter cell to which this Molecule belongs */
      next_cell_index = daughter_index[ random(daughter_index.size()) ];
    }

    /* Correct the cell_index of this Molecule */
    iter->set_cell_index(next_cell_index);

    /* Add this Molecule to the next-generation Cell. In ver005, I made
       Cell::add_mol() thread-safe. */
    cellpop_next.add_mol(next_cell_index, &(*iter));
  }

  /* Assign the parent_index of next-generation Cells */
  for(std::size_t curr_index=0, curr_end=cellpop_curr.get_n_alive_cells(); curr_index<curr_end; ++curr_index){
    for(auto daughter_iter=cellpop_curr.get_daughter_index(curr_index).begin(), daughter_end=cellpop_curr.get_daughter_index(curr_index).end(); daughter_iter!=daughter_end; ++daughter_iter){
      cellpop_next.set_parent_index(*daughter_iter, curr_index);
    }
  }
}

class System::DivideCell {
  MolVec& molvec_next;
  const CellPop& cellpop_curr;
  CellPop& cellpop_next;
  RandomTBB2::reference random;
  
public:
  void operator()(const tbb::blocked_range<std::size_t>& range) const {

    /* Fisher-Wright process */
    std::size_t curr_cell_index, next_cell_index;
    for(std::size_t mol_index=range.begin(), end=range.end(); mol_index!=end; ++mol_index){
      Molecule& mol = molvec_next[mol_index];
      
      /* Get the index of the current-generation Cell whose Molecule gave
	 rise to this Molecule in the next generation */
      curr_cell_index = mol.get_cell_index();

      /* Choose which daughter Cell this Molecule belongs to */
      const std::vector<std::size_t>& daughter_index = cellpop_curr.get_daughter_index(curr_cell_index);

      /* No cell division */
      if(daughter_index.size() == 1){
	next_cell_index = daughter_index[0];
      }
      /* Cell division occurs */
      else{
	/* Choose randomly a daughter cell to which this Molecule belongs */
	next_cell_index = daughter_index[ random(daughter_index.size()) ];
      }

      /* Correct the cell_index of this Molecule */
      mol.set_cell_index(next_cell_index);

      /* Add this Molecule to the next-generation Cell. In ver001, I made
	 Cell::add_mol() thread-safe. */
      cellpop_next.add_mol(mol.get_cell_index(), &mol);
    }
  }
  
  DivideCell(MolVec& _molvec_next, const CellPop& _cellpop_curr, CellPop& _cellpop_next):
    molvec_next(_molvec_next),
    cellpop_curr(_cellpop_curr),
    cellpop_next(_cellpop_next),
    random(random_tbb_2.local())
  {}
};

void System::divide_cell_parallel()
{
  MolVec& molvec_next = *molvec_next_ptr;  
  CellPop& cellpop_curr = *cellpop_curr_ptr;
  CellPop& cellpop_next = *cellpop_next_ptr;

  /* Set the daughter_index of current-generation Cells */
  std::size_t n_nextgen_cells = cellpop_curr.set_daughter_cell_index();

  /* Make next-generation Cells */
  cellpop_next.reset(n_nextgen_cells);

  /* Determine the index of Cells to which the next-generation Molecules
     belong */
  tbb::parallel_for(tbb::blocked_range<std::size_t>(0, molvec_next.size()), DivideCell(molvec_next,cellpop_curr,cellpop_next), tbb::static_partitioner());

  /* Assign the parent_index of next-generation Cells. This information
     enables ancestor tracking. The information will be output by
     observer.cpp */
  for(std::size_t curr_index=0, curr_end=cellpop_curr.get_n_alive_cells(); curr_index<curr_end; ++curr_index){
    for(auto daughter_iter=cellpop_curr.get_daughter_index(curr_index).begin(), daughter_end=cellpop_curr.get_daughter_index(curr_index).end(); daughter_iter!=daughter_end; ++daughter_iter){
      cellpop_next.set_parent_index(*daughter_iter, curr_index);
    }
  }
}

bool System::check_consistency() const
{
  MolVec& molvec_curr = *molvec_curr_ptr;
  CellPop& cellpop_curr = *cellpop_curr_ptr;

  cellpop_curr.check_consistency();


  for(auto& elem : molvec_curr){
    bool has = cellpop_curr.has_molecule(elem.get_cell_index(), &(elem));

    if(!has){
      ParallelPrint::cerr("System::check_consistency() ERROR: Some Molecule is not found in the Cell where it should be found\n");
      return false;
    }
  }

  return true;
}
